@@ -27,7 +27,6 @@
 #[cfg(feature = "tls")]
 use async_native_tls::TlsConnector;
 
-use std::borrow::Cow;
 use tokio::io::{AsyncReadExt, AsyncWriteExt, Error};
 use tokio::net::TcpStream;
 
@@ -123,32 +122,28 @@ impl Client {
     /// # Errors
     /// Returns error if the client could not write to the stream.
     pub async fn identify(&mut self) -> Result<(), Error> {
-        self.write_command(commands::Command::CAP(commands::CapMode::LS))
-            .await?;
-        self.write_command(commands::Command::CAP(commands::CapMode::END))
-            .await?;
+        self.cap_mode(commands::CapMode::LS).await?;
+        self.cap_mode(commands::CapMode::END).await?;
 
-        self.write_command(commands::Command::USER(
+        self.user(
             self.config.username.clone(),
-            "*".into(),
-            "*".into(),
+            "*",
+            "*",
             self.config.username.clone(),
-        ))
+        )
         .await?;
 
         if let Some(nick) = self.config.nickname.clone() {
-            self.write_command(commands::Command::NICK(nick)).await?;
+            self.nick(&nick).await?;
         } else {
-            self.write_command(commands::Command::NICK(self.config.username.clone()))
-                .await?;
+            self.nick(&self.config.username.clone()).await?;
         }
 
         loop {
             if let Ok(ref command) = self.read().await {
                 match command {
                     commands::Command::PING(code) => {
-                        self.write_command(commands::Command::PONG(code.to_string()))
-                            .await?;
+                        self.pong(code).await?;
                     }
                     commands::Command::OTHER(line) => {
                         tracing::debug!("{}", line);
@@ -169,11 +164,13 @@ impl Client {
                 config.username
             }
         };
-        self.write_command(commands::Command::MODE(user, config.mode))
-            .await?;
+
+        if let Some(mode) = config.mode {
+            self.mode(&user, Some(&mode)).await?;
+        }
+
         for channel in &config.channels {
-            self.write_command(commands::Command::JOIN(channel.to_string()))
-                .await?;
+            self.join(channel).await?;
         }
 
         tracing::info!("circe has identified");
@@ -219,8 +216,8 @@ impl Client {
         if let Some(string) = self.read_string().await {
             let command = commands::Command::command_from_str(&string).await;
 
-            if let commands::Command::PONG(command) = command {
-                if let Err(_e) = self.write_command(commands::Command::PONG(command)).await {
+            if let commands::Command::PING(command) = command {
+                if let Err(_e) = self.pong(&command).await {
                     return Err(NoNewLines);
                 }
                 return Ok(commands::Command::PONG("".to_string()));
@@ -232,15 +229,9 @@ impl Client {
         Err(NoNewLines)
     }
 
-    async fn write(&mut self, data: &str) -> Result<(), Error> {
-        let formatted = {
-            let new = format!("{}\r\n", data);
-            Cow::Owned(new) as Cow<str>
-        };
-
-        tracing::debug!("{:?}", formatted);
-
-        self.stream.write_all(formatted.as_bytes()).await?;
+    async fn write(&mut self, data: String) -> Result<(), Error> {
+        tracing::debug!("{:?}", data);
+        self.stream.write_all(data.as_bytes()).await?;
 
         Ok(())
     }
@@ -256,6 +247,7 @@ impl Client {
     /// ```
     /// # Errors
     /// Returns error if the client could not write to the stream.
+    #[deprecated]
     pub async fn write_command(&mut self, command: commands::Command) -> Result<(), Error> {
         use commands::Command::{
             ADMIN, AWAY, CAP, INVITE, JOIN, LIST, MODE, NAMES, NICK, OPER, OTHER, PART, PASS, PING,
@@ -263,27 +255,26 @@ impl Client {
         };
         let computed = match command {
             ADMIN(target) => {
-                let formatted = format!("ADMIN {}", target);
-                Cow::Owned(formatted) as Cow<str>
+                format!("ADMIN {}\r\n", target)
             }
             AWAY(message) => {
-                let formatted = format!("AWAY {}", message);
-                Cow::Owned(formatted) as Cow<str>
+                format!("AWAY {}\r\n", message)
             }
             CAP(mode) => {
                 use commands::CapMode::{END, LS};
-                Cow::Borrowed(match mode {
-                    LS => "CAP LS 302",
-                    END => "CAP END",
-                }) as Cow<str>
+                let formatted = {
+                    match mode {
+                        LS => "CAP LS 302\r\n",
+                        END => "CAP END\r\n",
+                    }
+                };
+                formatted.to_string()
             }
             INVITE(username, channel) => {
-                let formatted = format!("INVITE {} {}", username, channel);
-                Cow::Owned(formatted) as Cow<str>
+                format!("INVITE {} {}\r\n", username, channel)
             }
             JOIN(channel) => {
-                let formatted = format!("JOIN {}", channel);
-                Cow::Owned(formatted) as Cow<str>
+                format!("JOIN {}\r\n", channel)
             }
             LIST(channel, server) => {
                 let mut formatted = "LIST".to_string();
@@ -293,35 +284,28 @@ impl Client {
                 if let Some(server) = server {
                     formatted.push_str(format!(" {}", server).as_str());
                 }
-                Cow::Owned(formatted) as Cow<str>
+                formatted.push_str("\r\n");
+                formatted
             }
             NAMES(channel, server) => {
-                let formatted = {
-                    if let Some(server) = server {
-                        format!("NAMES {} {}", channel, server)
-                    } else {
-                        format!("NAMES {}", channel)
-                    }
-                };
-                Cow::Owned(formatted) as Cow<str>
+                if let Some(server) = server {
+                    format!("NAMES {} {}\r\n", channel, server)
+                } else {
+                    format!("NAMES {}\r\n", channel)
+                }
             }
             NICK(nickname) => {
-                let formatted = format!("NICK {}", nickname);
-                Cow::Owned(formatted) as Cow<str>
+                format!("NICK {}\r\n", nickname)
             }
             MODE(target, mode) => {
-                let formatted = {
-                    if let Some(mode) = mode {
-                        format!("MODE {} {}", target, mode)
-                    } else {
-                        format!("MODE {}", target)
-                    }
-                };
-                Cow::Owned(formatted) as Cow<str>
+                if let Some(mode) = mode {
+                    format!("MODE {} {}\r\n", target, mode)
+                } else {
+                    format!("MODE {}\r\n", target)
+                }
             }
             OPER(nick, password) => {
-                let formatted = format!("OPER {} {}", nick, password);
-                Cow::Owned(formatted) as Cow<str>
+                format!("OPER {} {}\r\n", nick, password)
             }
             OTHER(_) => {
                 return Err(Error::new(
@@ -330,46 +314,36 @@ impl Client {
                 ));
             }
             PART(target) => {
-                let formatted = format!("PART {}", target);
-                Cow::Owned(formatted) as Cow<str>
+                format!("PART {}\r\n", target)
             }
             PASS(password) => {
-                let formatted = format!("PASS {}", password);
-                Cow::Owned(formatted) as Cow<str>
+                format!("PASS {}\r\n", password)
             }
             PING(target) => {
-                let formatted = format!("PING {}", target);
-                Cow::Owned(formatted) as Cow<str>
+                format!("PING {}\r\n", target)
             }
             PONG(code) => {
-                let formatted = format!("PONG {}", code);
-                Cow::Owned(formatted) as Cow<str>
+                format!("PONG {}\r\n", code)
             }
             PRIVMSG(_, target, message) => {
-                let formatted = format!("PRIVMSG {} {}", target, message);
-                Cow::Owned(formatted) as Cow<str>
+                format!("PRIVMSG {} {}\r\n", target, message)
             }
             QUIT(message) => {
-                let formatted = format!("QUIT :{}", message);
-                Cow::Owned(formatted) as Cow<str>
+                format!("QUIT :{}\r\n", message)
             }
             TOPIC(channel, topic) => {
-                let formatted = {
-                    if let Some(topic) = topic {
-                        format!("TOPIC {} :{}", channel, topic)
-                    } else {
-                        format!("TOPIC {}", channel)
-                    }
-                };
-                Cow::Owned(formatted) as Cow<str>
+                if let Some(topic) = topic {
+                    format!("TOPIC {} :{}\r\n", channel, topic)
+                } else {
+                    format!("TOPIC {}\r\n", channel)
+                }
             }
             USER(username, s1, s2, realname) => {
-                let formatted = format!("USER {} {} {} :{}", username, s1, s2, realname);
-                Cow::Owned(formatted) as Cow<str>
+                format!("USER {} {} {} :{}\r\n", username, s1, s2, realname)
             }
         };
 
-        self.write(&computed).await?;
+        self.write(computed).await?;
         Ok(())
     }
 
@@ -385,8 +359,7 @@ impl Client {
     /// # Errors
     /// Returns error if the client could not write to the stream.
     pub async fn admin(&mut self, target: &str) -> Result<(), Error> {
-        self.write_command(commands::Command::ADMIN(target.to_string()))
-            .await?;
+        self.write(format!("ADMIN {}\r\n", target)).await?;
         Ok(())
     }
 
@@ -400,27 +373,17 @@ impl Client {
     /// # Errors
     /// Returns error if the client could not write to the stream.
     pub async fn away(&mut self, message: &str) -> Result<(), Error> {
-        self.write_command(commands::Command::AWAY(message.to_string()))
-            .await?;
+        self.write(format!("AWAY {}\r\n", message)).await?;
         Ok(())
     }
 
-    /// Helper function for sending PRIVMSGs.
-    /// ```no_run
-    /// # use circe::*;
-    /// # let mut client = Client::new(Default::default()).await?;
-    /// client.privmsg("#main", "Hello").await?;
-    /// # Ok::<(), tokio::io::Error>(())
-    /// ```
-    /// # Errors
-    /// Returns error if the client could not write to the stream.
-    pub async fn privmsg(&mut self, channel: &str, message: &str) -> Result<(), Error> {
-        self.write_command(commands::Command::PRIVMSG(
-            String::from(""),
-            channel.to_string(),
-            message.to_string(),
-        ))
-        .await?;
+    #[doc(hidden)]
+    pub async fn cap_mode(&mut self, mode: commands::CapMode) -> Result<(), Error> {
+        let cap_mode = match mode {
+            commands::CapMode::LS => "CAP LS 302\r\n",
+            commands::CapMode::END => "CAP END\r\n",
+        };
+        self.write(cap_mode.to_string()).await?;
         Ok(())
     }
 
@@ -434,11 +397,8 @@ impl Client {
     /// # Errors
     /// Returns error if the client could not write to the stream.
     pub async fn invite(&mut self, username: &str, channel: &str) -> Result<(), Error> {
-        self.write_command(commands::Command::INVITE(
-            username.to_string(),
-            channel.to_string(),
-        ))
-        .await?;
+        self.write(format!("INVITE {} {}\r\n", username, channel))
+            .await?;
         Ok(())
     }
 
@@ -452,8 +412,7 @@ impl Client {
     /// # Errors
     /// Returns error if the client could not write to the stream.
     pub async fn join(&mut self, channel: &str) -> Result<(), Error> {
-        self.write_command(commands::Command::JOIN(channel.to_string()))
-            .await?;
+        self.write(format!("JOIN {}\r\n", channel)).await?;
         Ok(())
     }
 
@@ -467,51 +426,15 @@ impl Client {
     /// # Errors
     /// Returns error if the client could not write to the stream.
     pub async fn list(&mut self, channel: Option<&str>, server: Option<&str>) -> Result<(), Error> {
-        let channel_config = channel.map(std::string::ToString::to_string);
-        let server_config = server.map(std::string::ToString::to_string);
-        self.write_command(commands::Command::LIST(channel_config, server_config))
-            .await?;
-        Ok(())
-    }
-
-    /// Helper function for getting all nicknames in a channel
-    /// ```no_run
-    /// # use circe::*;
-    /// # let mut client = Client::new(Default::default()).await?;
-    /// client.names("#main,#circe", None).await?;
-    /// # Ok::<(), tokio::io::Error>(())
-    /// ```
-    /// # Errors
-    /// Returns error if the client could not write to the stream.
-    pub async fn names(&mut self, channel: &str, server: Option<&str>) -> Result<(), Error> {
-        if let Some(server) = server {
-            self.write_command(commands::Command::NAMES(
-                channel.to_string(),
-                Some(server.to_string()),
-            ))
-            .await?;
-        } else {
-            self.write_command(commands::Command::NAMES(channel.to_string(), None))
-                .await?;
+        let mut formatted = "LIST".to_string();
+        if let Some(channel) = channel {
+            formatted.push_str(format!(" {}", channel).as_str());
         }
-        Ok(())
-    }
-
-    /// Helper function to try to register as a channel operator
-    /// ```no_run
-    /// # use circe::*;
-    /// # let mut client = Client::new(Default::default()).await?;
-    /// client.oper("username", "password").await?;
-    /// # Ok::<(), tokio::io::Error>(())
-    /// ```
-    /// # Errors
-    /// Returns error if the client could not write to the stream.
-    pub async fn oper(&mut self, username: &str, password: &str) -> Result<(), Error> {
-        self.write_command(commands::Command::OPER(
-            username.to_string(),
-            password.to_string(),
-        ))
-        .await?;
+        if let Some(server) = server {
+            formatted.push_str(format!(" {}", server).as_str());
+        }
+        formatted.push_str("\r\n");
+        self.write(formatted).await?;
         Ok(())
     }
 
@@ -526,18 +449,61 @@ impl Client {
     /// Returns error if the client could not write to the stream.
     pub async fn mode(&mut self, target: &str, mode: Option<&str>) -> Result<(), Error> {
         if let Some(mode) = mode {
-            self.write_command(commands::Command::MODE(
-                target.to_string(),
-                Some(mode.to_string()),
-            ))
-            .await?;
+            self.write(format!("MODE {} {}\r\n", target, mode,)).await?;
         } else {
-            self.write_command(commands::Command::MODE(target.to_string(), None))
-                .await?;
+            self.write(format!("MODE {}\r\n", target)).await?;
         }
         Ok(())
     }
 
+    /// Helper function for getting all nicknames in a channel
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Default::default()).await?;
+    /// client.names("#main,#circe", None).await?;
+    /// # Ok::<(), tokio::io::Error>(())
+    /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
+    pub async fn names(&mut self, channel: &str, server: Option<&str>) -> Result<(), Error> {
+        if let Some(server) = server {
+            self.write(format!("NAMES {} {}\r\n", channel, server))
+                .await?;
+        } else {
+            self.write(format!("NAMES {}\r\n", channel)).await?;
+        }
+        Ok(())
+    }
+
+    /// Helper function for changeing your nick
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Default::default()).await?;
+    /// client.nick("Not circe").await?;
+    /// # Ok::<(), tokio::io::Error>(())
+    /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
+    pub async fn nick(&mut self, nickname: &str) -> Result<(), Error> {
+        self.write(format!("NICK {}\r\n", nickname)).await?;
+        Ok(())
+    }
+
+    /// Helper function to try to register as a channel operator
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Default::default()).await?;
+    /// client.oper("username", "password").await?;
+    /// # Ok::<(), tokio::io::Error>(())
+    /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
+    pub async fn oper(&mut self, username: &str, password: &str) -> Result<(), Error> {
+        self.write(format!("OPER {} {}\r\n", username, password))
+            .await?;
+        Ok(())
+    }
+
     /// Helper function for leaving channels.
     /// ```no_run
     /// # use circe::*;
@@ -547,32 +513,55 @@ impl Client {
     /// ```
     /// # Errors
     /// Returns error if the client could not write to the stream.
-    pub async fn part(&mut self, target: &str) -> Result<(), Error> {
-        self.write_command(commands::Command::PART(target.to_string()))
-            .await?;
+    pub async fn part(&mut self, channel: &str) -> Result<(), Error> {
+        self.write(format!("PART {}\r\n", channel)).await?;
         Ok(())
     }
 
-    /// Helper function for setting or getting the topic of a channel
+    #[doc(hidden)]
+    pub async fn pass(&mut self, password: &str) -> Result<(), Error> {
+        self.write(format!("PASS {}\r\n", password)).await?;
+        Ok(())
+    }
+
+    /// Tests the presence of a connection.
     /// ```no_run
     /// # use circe::*;
     /// # let mut client = Client::new(Default::default()).await?;
-    /// client.topic("#main", Some("main channel")).await?;
+    /// client.ping("libera.chat", None).await?;
     /// # Ok::<(), tokio::io::Error>(())
     /// ```
     /// # Errors
     /// Returns error if the client could not write to the stream.
-    pub async fn topic(&mut self, channel: &str, topic: Option<&str>) -> Result<(), Error> {
-        if let Some(topic) = topic {
-            self.write_command(commands::Command::TOPIC(
-                channel.to_string(),
-                Some(topic.to_string()),
-            ))
-            .await?;
-        } else {
-            self.write_command(commands::Command::TOPIC(channel.to_string(), None))
+    pub async fn ping(&mut self, server1: &str, server2: Option<&str>) -> Result<(), Error> {
+        if let Some(server2) = server2 {
+            self.write(format!("PING {} {}\r\n", server1, server2))
                 .await?;
+        } else {
+            self.write(format!("PING {}\r\n", server1)).await?;
         }
+        // TODO look if we actually get a PONG back
+        Ok(())
+    }
+
+    #[doc(hidden)]
+    pub async fn pong(&mut self, ping: &str) -> Result<(), Error> {
+        self.write(format!("PONG {}\r\n", ping)).await?;
+        Ok(())
+    }
+
+    /// Helper function for sending PRIVMSGs.
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Default::default()).await?;
+    /// client.privmsg("#main", "Hello").await?;
+    /// # Ok::<(), tokio::io::Error>(())
+    /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
+    pub async fn privmsg(&mut self, channel: &str, message: &str) -> Result<(), Error> {
+        self.write(format!("PRIVMSG {} {}\r\n", channel, message))
+            .await?;
         Ok(())
     }
 
@@ -587,19 +576,50 @@ impl Client {
     /// Returns error if the client could not write to the stream.
     pub async fn quit(&mut self, message: Option<&str>) -> Result<(), Error> {
         if let Some(message) = message {
-            self.write_command(commands::Command::QUIT(message.to_string()))
-                .await?;
+            self.write(format!("QUIT :{}\r\n", message)).await?;
         } else {
-            self.write_command(commands::Command::QUIT(format!(
-                "circe {} (https://crates.io/crates/circe)",
+            self.write(format!(
+                "QUIT :circe {} (https://crates.io/crates/async-circe)\r\n",
                 env!("CARGO_PKG_VERSION")
-            )))
+            ))
             .await?;
         }
         tracing::info!("circe shutting down");
         self.stream.shutdown().await?;
         Ok(())
     }
+
+    /// Helper function for setting or getting the topic of a channel
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Default::default()).await?;
+    /// client.topic("#main", Some("main channel")).await?;
+    /// # Ok::<(), tokio::io::Error>(())
+    /// ```
+    /// # Errors
+    /// Returns error if the client could not write to the stream.
+    pub async fn topic(&mut self, channel: &str, topic: Option<&str>) -> Result<(), Error> {
+        if let Some(topic) = topic {
+            self.write(format!("TOPIC {} :{}\r\n", channel, topic))
+                .await?;
+        } else {
+            self.write(format!("TOPIC {}\r\n", channel)).await?;
+        }
+        Ok(())
+    }
+
+    #[doc(hidden)]
+    pub async fn user(
+        &mut self,
+        username: String,
+        s1: &str,
+        s2: &str,
+        realname: String,
+    ) -> Result<(), Error> {
+        self.write(format!("USER {} {} {} :{}\r\n", username, s1, s2, realname))
+            .await?;
+        Ok(())
+    }
 }
 
 impl Config {
